# Copyright 2020, 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Tuan Chien, Aniek Roelofs

"""
File utility functions
"""

import codecs
import gzip
import hashlib
import io
import logging
import os
import re
import shutil
import subprocess
from functools import partial
from pathlib import Path
from subprocess import Popen
from typing import List

import json_lines
import jsonlines
import numpy as np
import pandas as pd
from _hashlib import HASH
from google_crc32c import Checksum as Crc32cChecksum
from observatory.platform.utils.proc_utils import wait_for_process


def list_files(path: str, regex: str = None) -> List[str]:
    """Recursively list files on a path. Optionally only return files that match a regular expression.
    :param path: the path to recursively list files on.
    :param regex: a regular expression, if a file matches, then it is included.
    :return: a list of full paths to the files.
    """

    paths = []
    for root, dirs, files in os.walk(path):
        for file in files:
            if regex is None or re.match(regex, file):
                paths.append(os.path.join(root, file))
    return paths


def get_hasher_(algorithm: str) -> HASH:
    """Get the hasher for the specified algorithm.

    :param algorithm: Algorithm name.  See https://docs.python.org/3.8/library/hashlib.html for supported names.
    :return: Hasher object.
    """

    if algorithm == "md5":
        return hashlib.md5()
    elif algorithm == "sha256":
        return hashlib.sha256()
    elif algorithm == "sha512":
        return hashlib.sha512()

    raise Exception(f"get_hasher_ unsupported algorithm: {algorithm}")


def get_file_hash(*, file_path: str, algorithm: str = "md5") -> str:
    """Get the hash string of the file.

    :param file_path: File to hash.
    :param algorithm: Hashing algorithm to use.
    :return: Hash string.
    """

    hasher = get_hasher_(algorithm)

    BUFFER_SIZE = 2 ** 16  # 64 KiB
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(BUFFER_SIZE), b""):
            hasher.update(chunk)

    hash_code = hasher.hexdigest()
    return hash_code


def validate_file_hash(*, file_path: str, expected_hash: str, algorithm="md5") -> bool:
    """Check whether a file has the correct hash string.

    :param file_path: File to check.
    :param expected_hash: Expected hash string.
    :param algorithm: Hashing algorithm to use.
    :return: Whether the hash is valid.
    """

    computed_hash = get_file_hash(file_path=file_path, algorithm=algorithm)
    return computed_hash == expected_hash


def gzip_file_crc(file_path: str) -> str:
    """Get the crc of a gzip file.

    :param file_path: the path to the file.
    :return: the crc.
    """

    proc: Popen = subprocess.Popen(["gzip", "-vl", file_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output, error = wait_for_process(proc)
    return output.splitlines()[1].split(" ")[1].strip()


def hex_to_base64_str(hex_str: bytes) -> str:
    """Covert a hexadecimal string into a base64 encoded string. Removes trailing newline character.

    :param hex_str: the hexadecimal encoded string.
    :return: the base64 encoded string.
    """

    string = codecs.decode(hex_str, "hex")
    base64 = codecs.encode(string, "base64")
    return base64.decode("utf8").rstrip("\n")


def crc32c_base64_hash(file_path: str, chunk_size: int = 8 * 1024) -> str:
    """Create a base64 crc32c checksum of a file.

    :param file_path: the path to the file.
    :param chunk_size: the size of each chunk to check.
    :return: the checksum.
    """

    hash_alg = Crc32cChecksum()

    with open(file_path, "rb") as f:
        chunk = f.read(chunk_size)
        while chunk:
            hash_alg.update(chunk)
            chunk = f.read(chunk_size)
    return hex_to_base64_str(hash_alg.hexdigest())


def load_csv(file_path: str):
    """Return a CSV file as a list of dictionaries. It will check if the file is gzipped and unzip it if so.

    :param file_path: the path to the CSV file.
    :return: a list.
    """

    return list(yield_csv(file_path))


def is_gzip(file_path: str) -> bool:
    """Return whether a file is a gzip file or not

    :param file_path: the path to the file.
    :return: whether the file is a gzip file or not.
    """

    is_gzip_ = False
    with gzip.open(file_path, "rb") as f:
        try:
            f.read(1)
            is_gzip_ = True
        except OSError:
            pass

    return is_gzip_


def yield_csv(file_path: str):
    """Yield each row of a CSV file as a dictionary. It will check if the file is gzipped and unzip it if so.

    :param file_path: the path to the CSV file.
    :return: a generator.
    """

    if is_gzip(file_path):
        func = partial(gzip.open, file_path, mode="rb")
    else:
        func = partial(open, file_path, mode="r")

    with func() as f:
        df = pd.read_csv(f)
        df = df.replace({np.nan: None})
        for index, row in df.iterrows():
            yield row.to_dict()


def load_jsonl(file_path: str):
    """Return all rows of a JSON lines file as a list of dictionaries. If the file
    is gz compressed then it will be extracted.

    :param file_path: the path to the JSON lines file.
    :return: a list.
    """

    return list(yield_jsonl(file_path))


def yield_jsonl(file_path: str):
    """Return or yield row of a JSON lines file as a dictionary. If the file
    is gz compressed then it will be extracted.

    :param file_path: the path to the JSON lines file.
    :return: generator.
    """

    with json_lines.open(file_path) as file:
        for row in file:
            yield row


def list_to_jsonl_gz(file_path: str, list_of_dicts: List[dict]):
    """Takes a list of dictionaries and writes this to a gzipped jsonl file.
    :param file_path: Path to the .jsonl.gz file
    :param list_of_dicts: A list containing dictionaries that can be written out with jsonlines
    :return: None.
    """
    with io.BytesIO() as bytes_io:
        with gzip.GzipFile(fileobj=bytes_io, mode="w") as gzip_file:
            with jsonlines.Writer(gzip_file) as writer:
                writer.write_all(list_of_dicts)

        with open(file_path, "wb") as jsonl_gzip_file:
            jsonl_gzip_file.write(bytes_io.getvalue())


def load_file(file_name: str, modes="r"):
    """Load a file.

    :param file_name: file to load.
    :param modes: File open modes. Defaults to 'r'
    :return: contents of file.
    """

    with open(file_name, modes) as f:
        return f.read()


def write_to_file(record, file_name: str):
    """Write a structure to file.

    :param record: Structure to write.
    :param file_name: File name to write to.
    """

    directory = os.path.dirname(file_name)
    Path(directory).mkdir(parents=True, exist_ok=True)

    with open(file_name, "w") as f:
        f.write(record)


def gunzip_files(*, file_list: List[str], output_dir: str = None):
    """Gunzip the list of files.

    :param file_list: List of files to unzip.
    :param output_dir: Optional output directory.
    """

    for file_path in file_list:
        if file_path[-3:] != ".gz":  # Skip files without .gz extension
            continue

        logging.info(f"Unzipping file {file_path}")

        if output_dir is None:
            output_dir = os.path.dirname(file_path)

        filename = os.path.basename(file_path)
        filename = filename[:-3]  # Strip .gz extension
        dst = os.path.join(output_dir, filename)

        with gzip.open(file_path, "rb") as f_in:
            with open(dst, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)


def find_replace_file(*, src: str, dst: str, pattern: str, replacement: str):
    """Find an expression (can be a regex) in lines of a text file and replace it with a replacement string.
    You can optionally save the file in a different location.

    :param src: Input file.
    :param dst: Destination file.
    :param pattern: Expression to search for.
    :param replacement: Replacement string.
    """

    with open(src, "r") as f_in:
        with open(dst, "w") as f_out:
            for line in f_in:
                output = re.sub(pattern=pattern, repl=replacement, string=line)
                f_out.write(output)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Tuan Chien

import json
import logging
from datetime import datetime
from typing import Optional, Tuple, Union

import pendulum
from airflow.models.dag import DAG
from airflow.models.dagbag import DagBag
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.models.variable import Variable
from airflow.utils.state import DagRunState
from croniter import croniter

from observatory.platform.observatory_config import (
    TerraformConfig,
    VirtualMachine,
)
from observatory.platform.terraform.terraform_api import TerraformApi, TerraformVariable
from observatory.platform.utils.airflow_utils import (
    AirflowConns,
    AirflowVars,
    get_airflow_connection_password,
    send_slack_msg,
)
from observatory.platform.utils.workflow_utils import delete_old_xcoms
from observatory.platform.workflows.workflow import Workflow


class TerraformRelease:
    XCOM_START_TIME_VM = "start_time_vm"
    XCOM_PREV_START_TIME_VM = "prev_start_time_vm"
    XCOM_TERRAFORM_RUN_ID = "terraform_run_id"
    TERRAFORM_CREATE_VM_KEY = "airflow_worker_vm"
    TARGET_ADDRS = "module.airflow_worker_vm"  # Name of module in terraform configuration that will be targeted

    @property
    def terraform_api(self) -> TerraformApi:
        """Construct a TerraformApi object from the Airflow connection.

        :return: TerraformApi object.
        """

        token = get_airflow_connection_password(AirflowConns.TERRAFORM)
        return TerraformApi(token)

    @property
    def workspace_id(self) -> str:
        """Uses terraform API and workspace name to get the id of this workspace.

        :return: workspace id
        """

        organization = Variable.get(AirflowVars.TERRAFORM_ORGANIZATION)
        environment = Variable.get(AirflowVars.ENVIRONMENT)
        workspace = TerraformConfig.WORKSPACE_PREFIX + environment
        workspace_id = self.terraform_api.workspace_id(organization, workspace)
        return workspace_id

    def get_vm_info(self) -> Tuple[Optional[VirtualMachine], Optional[TerraformVariable]]:
        """Get the VirtualMachine data object, and TerraformVariable object for airflow_worker_vm.

        :return VirtualMachine and TerraformVariable objects.
        """

        variables = self.terraform_api.list_workspace_variables(self.workspace_id)

        for var in variables:
            if var.key == TerraformRelease.TERRAFORM_CREATE_VM_KEY:
                return VirtualMachine.from_hcl(var.value), var

        return None, None

    def update_terraform_vm_create_variable(self, value: bool):
        """Update the Terraform VM create flag.

        :param value: New value to set.
        """

        vm, vm_var = self.get_vm_info()
        vm.create = value
        logging.info(f"vm.create: {vm.create}")
        vm_var.value = vm.to_hcl()

        self.terraform_api.update_workspace_variable(vm_var, self.workspace_id)

    def create_terraform_run(self, *, dag_id: str, start_date: pendulum.DateTime) -> str:
        """Create a Terraform run and return the run ID.

        :param dag_id: DAG ID.
        :param start_date: Task instance start date.
        :return Terraform run ID.
        """

        message = f'Triggered from airflow DAG "{dag_id}" at {start_date}'
        run_id = self.terraform_api.create_run(self.workspace_id, TerraformRelease.TARGET_ADDRS, message)
        logging.info(f"Terraform run_id: {run_id}")

        return run_id

    def check_terraform_run_status(
        self, *, ti: TaskInstance, execution_date: pendulum.DateTime, project_id: str, run_id: str
    ):
        """Retrieve the terraform run status until it is in a finished state, either successful or errored. See
        https://www.terraform.io/docs/cloud/api/run.html for possible run_status values.
        If the run status is not successful and the environment isn't develop a warning message will be sent to a slack
        channel.

        :param ti: Task instance.
        :param execution_date: DagRun execution date.
        :param project_id: The google cloud project id that will be displayed in the slack message
        :param run_id: The run id of the Terraform run
        :return: None
        """

        run_status = None
        while run_status not in [
            "planned_and_finished",
            "applied",
            "errored",
            "discarded",
            "canceled",
            "force_canceled",
        ]:
            run_details = self.terraform_api.get_run_details(run_id)
            run_status = run_details["data"]["attributes"]["status"]

        logging.info(f"Run status: {run_status}")
        comments = f"Terraform run status: {run_status}"
        logging.info(f'Sending slack notification: "{comments}"')
        send_slack_msg(ti=ti, execution_date=execution_date, comments=comments, project_id=project_id)


def parse_datetime(dt: str) -> Optional[pendulum.DateTime]:
    """Try to parse datetime using pendulum.parse. Do not try to parse None.

    :param dt: Datetime string.
    :return: Datetime object, or None if failed.
    """

    if dt is None:
        return None

    return pendulum.parse(dt)


class VmCreateWorkflow(Workflow):
    """Workflow to spin up an Airflow worker VM (with Terraform)."""

    DAG_ID = "vm_create"

    def __init__(
        self,
        *,
        start_date: pendulum.DateTime = pendulum.datetime(2020, 7, 1),
        schedule_interval: str = "@weekly",
    ):
        """Construct the workflow.

        :param start_date: Start date for the DAG.
        :param schedule_interval: Schedule interval for the DAG.
        """

        airflow_vars = [
            AirflowVars.PROJECT_ID,
            AirflowVars.TERRAFORM_ORGANIZATION,
            AirflowVars.ENVIRONMENT,
        ]
        airflow_conns = [AirflowConns.TERRAFORM]

        super().__init__(
            dag_id=VmCreateWorkflow.DAG_ID,
            start_date=start_date,
            schedule_interval=schedule_interval,
            catchup=False,
            max_active_runs=1,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
        )

        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.check_vm_state)
        self.add_task(self.update_terraform_variable)
        self.add_task(self.run_terraform)
        self.add_task(self.check_run_status)
        self.add_task(self.cleanup, trigger_rule="none_failed")

    def make_release(self, **kwargs) -> TerraformRelease:
        """Required for Workflow class.

        :param kwargs: Unused.
        :return: TerraformRelease.
        """

        return TerraformRelease()

    def check_vm_state(self, **kwargs) -> bool:
        """Checks if VM is running. Proceed only if VM is not already running.

        :param kwargs: Unused.
        :return: Whether to continue.
        """

        vm, _ = TerraformRelease().get_vm_info()
        logging.info(f"VM is on: {vm.create}")
        return not vm.create

    def update_terraform_variable(self, release: TerraformRelease, **kwargs):
        """Update Terraform variable for VM to running state.

        :param release: TerraformRelease object.
        :param kwargs: Unused.
        """

        release.update_terraform_vm_create_variable(True)

    def run_terraform(self, release: TerraformRelease, **kwargs):
        """Runs terraform configuration. The current task start time, previous task start time, and Terraform run ID will be pushed to XComs.

        :param release: TerraformRelease object.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        """

        ti: TaskInstance = kwargs["ti"]

        prev_start_time_vm = ti.xcom_pull(key=release.XCOM_START_TIME_VM, include_prior_dates=True)
        ti.xcom_push(release.XCOM_PREV_START_TIME_VM, prev_start_time_vm)
        ti.xcom_push(release.XCOM_START_TIME_VM, ti.start_date.isoformat())

        run_id = release.create_terraform_run(dag_id=self.dag_id, start_date=ti.start_date)
        ti.xcom_push(release.XCOM_TERRAFORM_RUN_ID, run_id)

    def check_run_status(self, release: TerraformRelease, **kwargs):
        """Retrieve the terraform run status until it is in a finished state, either successful or errored. See
        https://www.terraform.io/docs/cloud/api/run.html for possible run_status values.
        If the run status is not successful and the environment isn't develop a warning message will be sent to a slack
        channel.

        :param release: TerraformRelease object.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        """

        ti: TaskInstance = kwargs["ti"]
        execution_date = kwargs["execution_date"]

        run_id = ti.xcom_pull(key=release.XCOM_TERRAFORM_RUN_ID, task_ids=self.run_terraform.__name__)
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        release.check_terraform_run_status(ti=ti, execution_date=execution_date, project_id=project_id, run_id=run_id)

    def cleanup(self, release: TerraformRelease, **kwargs):
        """Delete stale XCom messages.

        :param release: TerraformRelease object.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        """

        execution_date = kwargs["execution_date"]
        delete_old_xcoms(dag_id=self.dag_id, execution_date=execution_date, retention_days=15)


class VmDestroyWorkflow(Workflow):
    """Workflow to teardown an Airflow worker VM (with Terraform)."""

    DAG_ID = "vm_destroy"

    VM_RUNTIME_H_WARNING = 20  # Uptime threshold for VM before slack notification
    WARNING_FREQUENCY_H = 5  # Frequency of slack notifications

    XCOM_WARNING_TIME = "last_warning_time"
    XCOM_DESTROY_TIME_VM = "destroy_time_vm"

    def __init__(
        self,
        *,
        start_date: pendulum.DateTime = pendulum.datetime(2020, 1, 1),
        schedule_interval: str = "*/10 * * * *",
    ):
        """Construct the workflow.

        :param start_date: Start date for the DAG.
        :param schedule_interval: Schedule interval for the DAG.
        """

        airflow_vars = [
            AirflowVars.PROJECT_ID,
            AirflowVars.TERRAFORM_ORGANIZATION,
            AirflowVars.VM_DAGS_WATCH_LIST,
            AirflowVars.ENVIRONMENT,
        ]
        airflow_conns = [AirflowConns.TERRAFORM]

        super().__init__(
            dag_id=VmDestroyWorkflow.DAG_ID,
            start_date=start_date,
            schedule_interval=schedule_interval,
            catchup=False,
            max_active_runs=1,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
        )

        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.check_vm_state)
        self.add_setup_task(self.check_dags_status)
        self.add_task(self.update_terraform_variable)
        self.add_task(self.run_terraform)
        self.add_task(self.check_run_status)
        self.add_task(self.cleanup, trigger_rule="none_failed")

    def make_release(self, **kwargs) -> TerraformRelease:
        """Required for Workflow class.

        :param kwargs: Unused.
        :return: TerraformRelease object.
        """

        return TerraformRelease()

    def check_vm_state(self, **kwargs) -> bool:
        """Checks if VM is running. Proceed only if VM is running.

        :param kwargs: Unused.
        :return: Whether to continue.
        """

        vm, _ = TerraformRelease().get_vm_info()
        logging.info(f"VM is on: {vm.create}")
        return vm.create

    def check_dags_status(self, **kwargs):
        """Check if all expected runs for the DAGs in the watchlist are successful. If they are the task, then proceed, otherwise check how long the VM has run for, and skip the rest of the workflow.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: id of task which should be executed next
        """

        ti: TaskInstance = kwargs["ti"]
        destroy_worker_vm = True

        release = TerraformRelease()

        prev_start_time_vm = ti.xcom_pull(
            key=release.XCOM_PREV_START_TIME_VM,
            task_ids=VmCreateWorkflow.run_terraform.__name__,
            dag_id=VmCreateWorkflow.DAG_ID,
            include_prior_dates=True,
        )
        prev_start_time_vm = parse_datetime(prev_start_time_vm)

        start_time_vm = ti.xcom_pull(
            key=release.XCOM_START_TIME_VM,
            task_ids=VmCreateWorkflow.run_terraform.__name__,
            dag_id=VmCreateWorkflow.DAG_ID,
            include_prior_dates=True,
        )
        start_time_vm = parse_datetime(start_time_vm)

        destroy_time_vm = ti.xcom_pull(
            key=VmDestroyWorkflow.XCOM_DESTROY_TIME_VM,
            task_ids=VmDestroyWorkflow.run_terraform.__name__,
            dag_id=VmDestroyWorkflow.DAG_ID,
            include_prior_dates=True,
        )
        destroy_time_vm = parse_datetime(destroy_time_vm)

        logging.info(
            f"prev_start_time_vm: {prev_start_time_vm}, start_time_vm: {start_time_vm}, "
            f"destroy_time_vm: {destroy_time_vm}\n"
        )

        # Load VM DAGs watch list
        vm_dags_watch_list_str = Variable.get(AirflowVars.VM_DAGS_WATCH_LIST)
        logging.info(f"vm_dags_watch_list_str str: {vm_dags_watch_list_str}")
        vm_dags_watch_list = json.loads(vm_dags_watch_list_str)
        logging.info(f"vm_dags_watch_list: {vm_dags_watch_list}")

        for dag_id in vm_dags_watch_list:
            dagbag = DagBag()
            dag = dagbag.get_dag(dag_id)
            logging.info(f"Dag id: {dag_id}")

            # vm turned on manually and never turned on before
            if not start_time_vm and not prev_start_time_vm:
                logging.warning("Both start_time_vm and prev_start_time_vm are None. Unsure whether to turn off DAG.")
                destroy_worker_vm = False
                break

            # returns last execution date of previous vm cycle or None if a DAG is running
            last_execution_prev = self._get_last_execution_prev(dag, dag_id, prev_start_time_vm)
            if not last_execution_prev:
                destroy_worker_vm = False
                break

            logging.info(f"Execution date of last DAG before prev_start_time_vm: {last_execution_prev}\n")

            if destroy_time_vm:
                if start_time_vm < destroy_time_vm:
                    # If the vm is on, but there's no start_time_vm it must have been turned on manually.
                    logging.warning(
                        "start_time_vm is before destroy_time_vm. Perhaps the vm was turned on "
                        "manually in between. This task will continue with the given start_time_vm."
                    )

            # get a backfill of all expected runs between last execution date of prev cycle and the time the vm was
            # started. create iterator starting at the latest planned schedule before start_time_vm
            cron_iter = croniter(dag.normalized_schedule_interval, dag.previous_schedule(start_time_vm))

            # the last_execution_current is expected 1 schedule interval before the DAGs 'previous_schedule',
            # because airflow won't trigger a DAG until 1 schedule interval after the 'execution_date'.
            last_execution_current = cron_iter.get_prev(datetime)

            # if DAG is not set to catchup any backfill, the only run date is the last one.
            if dag.catchup:
                execution_dates = dag.get_run_dates(last_execution_prev, last_execution_current)
            else:
                execution_dates = [last_execution_current]

            # for each execution date check if state is success. This can't be done in all_dag_runs above, because the
            # dag_run might not be in all_dag_runs yet, because it is not scheduled yet.
            destroy_worker_vm = self._check_success_runs(dag_id, execution_dates)
            if destroy_worker_vm is False:
                break

        logging.info(f"Destroying worker VM: {destroy_worker_vm}")

        # If not destroying vm, check VM runtime.
        if not destroy_worker_vm:
            self.check_runtime_vm(start_time_vm, **kwargs)

        return destroy_worker_vm

    def check_runtime_vm(self, start_time_vm: Optional[datetime], **kwargs):
        """Checks how long the VM has been turned on based on the xcom value from the terraform run task.
        A warning message will be sent in a slack channel if it has been on longer than the warning limit,
        the environment isn't develop and a message hasn't been sent already in the last x hours.

        :param start_time_vm: Start time of the vm
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: None
        """

        ti: TaskInstance = kwargs["ti"]
        last_warning_time = ti.xcom_pull(
            key=VmDestroyWorkflow.XCOM_WARNING_TIME,
            task_ids=ti.task_id,
            dag_id=VmDestroyWorkflow.DAG_ID,
            include_prior_dates=True,
        )
        last_warning_time = parse_datetime(last_warning_time)

        if start_time_vm:
            # calculate number of hours passed since start time vm and now
            hours_on = (ti.start_date - start_time_vm).total_seconds() / 3600
            logging.info(
                f"Start time VM: {start_time_vm}, hours passed since start time: {hours_on}, warning limit: "
                f"{VmDestroyWorkflow.VM_RUNTIME_H_WARNING}"
            )

            # check if a warning has been sent previously and if so, how many hours ago
            if last_warning_time:
                hours_since_warning = (ti.start_date - last_warning_time).total_seconds() / 3600
            else:
                hours_since_warning = None

            # check if the VM has been on longer than the limit
            if hours_on > VmDestroyWorkflow.VM_RUNTIME_H_WARNING:
                #  check if no warning was sent before or last time was longer ago than warning frequency
                if not hours_since_warning or hours_since_warning > VmDestroyWorkflow.WARNING_FREQUENCY_H:
                    comments = (
                        f"Worker VM has been on since {start_time_vm}. No. hours passed since then: "
                        f"{hours_on}."
                        f" Warning limit: {VmDestroyWorkflow.VM_RUNTIME_H_WARNING}H"
                    )
                    project_id = Variable.get(AirflowVars.PROJECT_ID)
                    execution_date = kwargs["execution_date"]
                    send_slack_msg(ti=ti, execution_date=execution_date, comments=comments, project_id=project_id)

                    ti.xcom_push(VmDestroyWorkflow.XCOM_WARNING_TIME, ti.start_date.isoformat())
        else:
            logging.info(f"Start time VM unknown.")

    def update_terraform_variable(self, release: TerraformRelease, **kwargs):
        """Update Terraform variable for VM to running state.

        :param release: TerraformRelease object.
        :param kwargs: Unused.
        """

        release.update_terraform_vm_create_variable(False)

    def run_terraform(self, release: TerraformRelease, **kwargs):
        """Runs terraform configuration. The current task start time, previous task start time, and Terraform run ID will be pushed to XComs.

        :param release: TerraformRelease object.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        """

        ti: TaskInstance = kwargs["ti"]
        ti.xcom_push(VmDestroyWorkflow.XCOM_DESTROY_TIME_VM, ti.start_date.isoformat())
        run_id = release.create_terraform_run(dag_id=self.dag_id, start_date=ti.start_date)
        ti.xcom_push(release.XCOM_TERRAFORM_RUN_ID, run_id)

    def check_run_status(self, release: TerraformRelease, **kwargs):
        """Retrieve the terraform run status until it is in a finished state, either successful or errored. See
        https://www.terraform.io/docs/cloud/api/run.html for possible run_status values.
        If the run status is not successful and the environment isn't develop a warning message will be sent to a slack
        channel.

        :param release: TerraformRelease object.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        """

        ti: TaskInstance = kwargs["ti"]
        execution_date = kwargs["execution_date"]

        run_id = ti.xcom_pull(key=release.XCOM_TERRAFORM_RUN_ID, task_ids=self.run_terraform.__name__)
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        release.check_terraform_run_status(ti=ti, execution_date=execution_date, project_id=project_id, run_id=run_id)

    def cleanup(self, release: TerraformRelease, **kwargs):
        """Delete stale XCom messages.

        :param release: TerraformRelease object.
        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        """

        execution_date = kwargs["execution_date"]
        delete_old_xcoms(dag_id=self.dag_id, execution_date=execution_date, retention_days=15)

    def _get_last_execution_prev(
        self, dag: DAG, dag_id: str, prev_start_time_vm: Union[datetime, None]
    ) -> Union[datetime, None]:
        """Find the execution date of the last DAG run before the previous time the VM was turned on.
        If there aren't any DAG runs before this time or the time is None (first/second time turning off VM) the
        execution date is set to the start_date of the DAG instead.

        If a DAG is currently running it will return None and the remaining tasks are skipped.

        :param dag: DAG object
        :param dag_id: the dag id
        :param prev_start_time_vm: previous time the VM was turned on
        :return: execution date or None
        """

        # Get execution date of the last run before previous start date
        all_dag_runs = DagRun.find(dag_id=dag_id)
        # sort dag runs by start datetime, newest first
        for dag_run in sorted(all_dag_runs, key=lambda x: x.start_date, reverse=True):
            if dag_run.state == "running":
                logging.info("DAG is currently running.")
                return None
            # None if first time running destroy
            if prev_start_time_vm:
                if pendulum.instance(dag_run.start_date) < prev_start_time_vm:
                    # get execution date of last run from when the VM was previously on
                    return dag_run.execution_date

        # No runs executed previously
        if prev_start_time_vm:
            logging.info("No DAG runs that started before prev_start_time_vm.")
        else:
            # First time running destroy_vm, no previous start date available
            logging.info("No prev_start_time_vm.")
        logging.info("Setting last execution date to start_date of DAG.")
        last_execution_prev = dag.default_args["start_date"]

        return last_execution_prev

    def _check_success_runs(self, dag_id: str, execution_dates: list) -> bool:
        """For each date in the execution dates it checks if a DAG run exists and if so if the state is set to success.

        Only if both of these are true for all dates it will return True.

        :param dag_id: the dag id
        :param execution_dates: list of execution dates
        :return: True or False
        """
        for date in execution_dates:
            dag_runs = DagRun.find(dag_id=dag_id, execution_date=date)
            if not dag_runs:
                logging.info(f"Expected dag run on {date} has not been scheduled yet")
                return False

            for dag_run in dag_runs:
                logging.info(
                    f"id: {dag_run.dag_id}, start date: {dag_run.start_date}, execution date: "
                    f"{dag_run.execution_date}, state: {dag_run.state}"
                )
                if dag_run.state != DagRunState.SUCCESS:
                    return False
        return True



# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs

import logging
import os
import time
from typing import Dict

import requests
from elasticsearch import Elasticsearch

from observatory.platform.docker.compose_runner import ComposeRunner, ProcessOutput
from observatory.platform.utils.config_utils import module_file_path


class ElasticEnvironment(ComposeRunner):
    HTTP_OK = 200

    def __init__(
        self,
        build_path: str,
        elastic_port: int = 9200,
        kibana_port: int = 5601,
        password: str = "observatory",
        wait: bool = True,
        wait_time_secs: int = 120,
    ):
        """Construct an Elasticsearch and Kibana environment.

        :param build_path: the path to the build directory.
        :param elastic_port: the Elastic port.
        :param kibana_port: the Kibana port.
        :param wait: whether to wait until Elastic and Kibana have started.
        :param wait_time_secs: the maximum wait time in seconds.
        """

        self.elastic_module_path = module_file_path("observatory.platform.elastic")
        self.wait = wait
        self.wait_time_secs = wait_time_secs
        self.elastic_port = elastic_port
        self.elastic_uri = f"http://localhost:{elastic_port}/"
        self.kibana_uri = f"http://localhost:{kibana_port}/"
        self.password = password
        self.build_path = build_path
        super().__init__(
            compose_template_path=os.path.join(self.elastic_module_path, "docker-compose.yml.jinja2"),
            build_path=build_path,
            compose_template_kwargs={"elastic_port": elastic_port, "kibana_port": kibana_port, "password": password},
            debug=True,
        )

        # Stop the awful unnecessary Elasticsearch connection warnings being logged
        logging.basicConfig()
        logging.getLogger().setLevel(logging.ERROR)

    def make_environment(self) -> Dict:
        """Make the environment when running the Docker Compose command.

        :return: the environment.
        """
        return os.environ.copy()

    def start(self) -> ProcessOutput:
        """Start the Elastic environment.

        :return: ProcessOutput.
        """

        self.stop()
        process_output = super().start()
        if self.wait:
            self.wait_until_started()
        time.sleep(20)
        return process_output

    def kibana_ping(self):
        """Check if Kibana has started or not.

        :return: whether Kibana has started or not.
        """

        try:
            # Have to call a specific API endpoint, not jus the base Kibana URI, as the base Kibana URI will return 200
            # before Kibana is actually ready.
            response = requests.get(f"{self.kibana_uri}/api/spaces/space")
            return response.status_code == self.HTTP_OK
        except (ConnectionResetError, requests.exceptions.ConnectionError):
            pass
        return False

    def wait_until_started(self):
        """Wait until Elastic and Kibana have started.

        :return: whether started or not.
        """
        es = Elasticsearch([self.elastic_uri])
        start = time.time()
        while True:
            try:
                elastic_found = es.ping()
            except OverflowError:
                elastic_found = False
            kibana_found = self.kibana_ping()
            services_found = elastic_found and kibana_found
            if services_found:
                break

            # Break out if time lasts too long
            elapsed = time.time() - start
            if elapsed > self.wait_time_secs:
                break

        return services_found



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

from observatory.platform.docker.platform_runner import PlatformRunner, HOST_UID, DEBUG
from observatory.platform.observatory_config import ObservatoryConfig
from observatory.platform.utils.url_utils import wait_for_url


class PlatformCommand(PlatformRunner):
    def __init__(self, config: ObservatoryConfig, host_uid: int = HOST_UID, debug: bool = DEBUG):
        """Create a PlatformCommand, which can be used to start and stop Observatory Platform instances.
        :param config: The path to the observatory config.
        :param host_uid: The user id of the host system. Used to set the user id in the Docker containers.
        :param debug: Print debugging information.
        """

        super().__init__(config=config, host_uid=host_uid, debug=debug)

    @property
    def ui_url(self) -> str:
        """Return the URL to Apache Airflow UI.
        :return: Apache Airflow UI URL.
        """

        return f"http://localhost:{self.config.observatory.airflow_ui_port}"

    def wait_for_airflow_ui(self, timeout: int = 60) -> bool:
        """Wait for the Apache Airflow UI to start.
        :param timeout: the number of seconds to wait before timing out.
        :return: whether connecting to the Apache Airflow UI was successful or not.
        """

        return wait_for_url(self.ui_url, timeout=timeout)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Author: Aniek Roelofs, Tuan Chien

"""
Airflow utility functions (independent of telescope or google cloud usage)
"""

import logging
from typing import List, Union

import pendulum
import validators
from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.models import TaskInstance, Variable
from airflow.providers.slack.operators.slack_webhook import SlackWebhookHook


class AirflowVars:
    """Common Airflow Variable names used with the Observatory Platform"""

    DATA_PATH = "data_path"
    ENVIRONMENT = "environment"
    PROJECT_ID = "project_id"
    DATA_LOCATION = "data_location"
    DOWNLOAD_BUCKET = "download_bucket"
    TRANSFORM_BUCKET = "transform_bucket"
    TERRAFORM_ORGANIZATION = "terraform_organization"
    DAGS_MODULE_NAMES = "dags_module_names"
    ORCID_BUCKET = "orcid_bucket"
    VM_DAGS_WATCH_LIST = "vm_dags_watch_list"
    BIGQUERY_DAILY_BYTES_LIMIT = "bigquery_daily_bytes_limit"


class AirflowConns:
    """Common Airflow Connection names used with the Observatory Platform"""

    CROSSREF = "crossref"
    TERRAFORM = "terraform"
    SLACK = "slack"
    GEOIP_LICENSE_KEY = "geoip_license_key"
    OAPEN_IRUS_UK_API = "oapen_irus_uk_api"
    OAPEN_IRUS_UK_LOGIN = "oapen_irus_uk_login"
    OAEBU_SERVICE_ACCOUNT = "oaebu_service_account"
    SFTP_SERVICE = "sftp_service"
    OBSERVATORY_API = "observatory_api"
    GMAIL_API = "gmail_api"
    ORCID = "orcid"


def change_task_log_level(new_levels: Union[List, int]) -> list:
    """Change the logging levels of all handlers for an airflow task.

    :param new_levels: New logging levels that all handlers will be set to
    :return: List of the old logging levels, can be used to restore logging levels.
    """
    logger = logging.getLogger("airflow.task")
    # stores logging levels
    old_levels = []
    for count, handler in enumerate(logger.handlers):
        old_levels.append(handler.level)
        if isinstance(new_levels, int):
            handler.setLevel(new_levels)
        else:
            handler.setLevel(new_levels[count])
    return old_levels


def check_variables(*variables):
    """Checks whether all given airflow variables exist.

    :param variables: name of airflow variable
    :return: True if all variables are valid
    """
    is_valid = True
    for name in variables:
        try:
            Variable.get(name)
        except KeyError:
            logging.error(f"Airflow variable '{name}' not set.")
            is_valid = False
    return is_valid


def check_connections(*connections):
    """Checks whether all given airflow connections exist.

    :param connections: name of airflow connection
    :return: True if all connections are valid
    """
    is_valid = True
    for name in connections:
        try:
            BaseHook.get_connection(name)
        except KeyError:
            logging.error(f"Airflow connection '{name}' not set.")
            is_valid = False
    return is_valid


def send_slack_msg(
    *,
    ti: TaskInstance,
    execution_date: pendulum.DateTime,
    comments: str = "",
    project_id: str = "?",
):
    """
    Send a slack message using the token in the slack airflow connection.
    :param ti: Task instance.
    :param execution_date: DagRun execution date.
    :param comments: Additional comments in slack message
    :param project_id: The google cloud project id that will be displayed in the slack message
    """

    message = """
    :red_circle: Task Alert.
    *Task*: {task}
    *Dag*: {dag}
    *Execution Time*: {exec_date}
    *Log Url*: {log_url}
    *Project id*: {project_id}
    *Comments*: {comments}
    """.format(
        task=ti.task_id,
        dag=ti.dag_id,
        exec_date=execution_date,
        log_url=ti.log_url,
        comments=comments,
        project_id=project_id,
    )
    slack_conn = BaseHook.get_connection(AirflowConns.SLACK)
    slack_hook = SlackWebhookHook(http_conn_id=slack_conn.conn_id, webhook_token=slack_conn.password, message=message)

    # http_hook outputs the secret token, suppressing logging 'info' by setting level to 'warning'
    old_levels = change_task_log_level(logging.WARNING)
    slack_hook.execute()
    # change back to previous levels
    change_task_log_level(old_levels)


def set_task_state(success: bool, task_id: str):
    """Update the state of the Airflow task.
    :param success: whether the task was successful or not.
    :param task_id: the task id.
    :return: None.
    """

    if success:
        logging.info(f"{task_id} success")
    else:
        msg_failed = f"{task_id} failed"
        logging.error(msg_failed)
        raise AirflowException(msg_failed)


def get_airflow_connection_url(conn_id: str) -> str:
    """Get the Airflow connection host, validate it is a valid url, and return it if it is, with a trailing /,
        otherwise throw an exception.  Assumes the connection_id exists.

    :param conn_id: Airflow connection id.
    :return: Connection URL with a trailing / added if necessary, or raise an exception if it is not a valid URL.
    """

    conn = BaseHook.get_connection(conn_id)
    url = conn.get_uri()

    if validators.url(url) != True:
        raise AirflowException(f"Airflow connection id {conn_id} does not have a valid url: {url}")

    if url[-1] != "/":
        url += "/"

    return url


def get_airflow_connection_login(conn_id: str) -> str:
    """Get the Airflow connection login. Assumes the connection_id exists.

    :param conn_id: Airflow connection id.
    :return: Connection login.
    """

    conn = BaseHook.get_connection(conn_id)
    return conn.login


def get_airflow_connection_password(conn_id: str) -> str:
    """Get the Airflow connection password. Assumes the connection_id exists.

    :param conn_id: Airflow connection id.
    :return: Connection password.
    """

    conn = BaseHook.get_connection(conn_id)
    password = conn.get_password()
    return password



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose, Tuan Chien

import json
import logging
import os
import pathlib
from typing import Dict, List, Tuple, Optional

import pendulum
from airflow.exceptions import AirflowException
from airflow.models import Variable
from google.cloud.bigquery import SourceFormat

from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.gc_utils import bigquery_sharded_table_id, create_bigquery_snapshot
from observatory.platform.utils.release_utils import (
    get_dataset_releases,
    get_datasets,
    get_latest_dataset_release,
    is_first_release,
)
from observatory.platform.utils.workflow_utils import (
    batch_blob_name,
    blob_name,
    bq_append_from_file,
    bq_append_from_partition,
    bq_delete_old,
    bq_load_ingestion_partition,
    delete_old_xcoms,
    table_ids_from_path,
    upload_files_from_list,
)
from observatory.platform.workflows.workflow import Release, Workflow


class StreamRelease(Release):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        end_date: pendulum.DateTime,
        first_release: bool,
        download_files_regex: str = None,
        extract_files_regex: str = None,
        transform_files_regex: str = None,
    ):
        """Construct a StreamRelease instance

        :param dag_id: the id of the DAG.
        :param start_date: the start_date of the release.
        :param end_date: the end_date of the release.
        :param first_release: whether this is the first release that is processed for this DAG
        :param download_files_regex: regex pattern that is used to find files in download folder
        :param extract_files_regex: regex pattern that is used to find files in extract folder
        :param transform_files_regex: regex pattern that is used to find files in transform folder
        """
        self.start_date = start_date
        self.end_date = end_date
        self.first_release = first_release
        release_id = self.start_date.strftime("%Y_%m_%d") + "-" + self.end_date.strftime("%Y_%m_%d")
        super().__init__(dag_id, release_id, download_files_regex, extract_files_regex, transform_files_regex)


class StreamTelescope(Workflow):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        schedule_interval: str,
        dataset_id: str,
        merge_partition_field: str,
        schema_folder: str,
        workflow_id: int,
        dataset_type_id: str,
        queue: str = "default",
        max_retries: int = 3,
        max_active_runs: int = 1,
        source_format: str = SourceFormat.NEWLINE_DELIMITED_JSON,
        schema_prefix: str = "",
        schema_version: str = None,
        load_bigquery_table_kwargs: Dict = None,
        dataset_description: str = "",
        table_descriptions: Dict[str, str] = None,
        batch_load: bool = False,
        airflow_vars: list = None,
        airflow_conns: list = None,
        tags: Optional[List[str]] = None,
        **kwargs,
    ):
        """Construct a StreamTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the dataset id.
        :param merge_partition_field: the BigQuery field used to match partitions for a merge
        :param schema_folder: the path to the SQL schema folder.
        :param workflow_id: Observatory API workflow id.
        :param dataset_type_id: Observatory API type_id for the top level dataset type.
        :param queue: the Airflow queue name.
        :param max_retries: the number of times to retry each task.
        :param max_active_runs: the maximum number of DAG runs that can be run at once.
        :param source_format: the format of the data to load into BigQuery.
        :param schema_prefix: the prefix used to find the schema path
        :param schema_version: the version used to find the schema path
        :param load_bigquery_table_kwargs: the customisation parameters for loading data into a BigQuery table.
        :param dataset_description: description for the BigQuery dataset.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions
        :param batch_load: whether all files in the transform folder are loaded into 1 table at once
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        :param airflow_conns: list of airflow connection keys, for each connection it is checked if it exists in airflow
        :param tags: Optional Airflow DAG tags to add.
        """

        # Set transform_bucket_name as required airflow variable
        if not airflow_vars:
            airflow_vars = []
        airflow_vars = list(set([AirflowVars.TRANSFORM_BUCKET] + airflow_vars))

        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            catchup=False,
            queue=queue,
            max_retries=max_retries,
            max_active_runs=max_active_runs,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
            workflow_id=workflow_id,
            dataset_type_id=dataset_type_id,
            tags=tags,
            **kwargs,
        )

        self.schema_prefix = schema_prefix
        self.schema_version = schema_version
        self.dataset_id = dataset_id
        self.source_format = source_format
        self.merge_partition_field = merge_partition_field
        self.schema_folder = schema_folder
        self.load_bigquery_table_kwargs = load_bigquery_table_kwargs if load_bigquery_table_kwargs else dict()
        self.dataset_description = dataset_description
        self.table_descriptions = table_descriptions if table_descriptions else dict()
        self.batch_load = batch_load

    def get_release_info(self, **kwargs) -> Tuple[pendulum.DateTime, pendulum.DateTime, bool]:
        """Return release information with the start and end date. [start date, end date) Includes start date, excludes end date.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """

        # Check if first release or not
        first_release = is_first_release(self.workflow_id)
        if first_release:
            # When first release, set start date to the start of the DAG
            # Can we get away with using data_interval_start instead?
            start_date = pendulum.instance(kwargs["dag"].default_args["start_date"]).start_of("day")
        else:
            datasets = get_datasets(workflow_id=self.workflow_id)
            try:
                top_level_dataset = [
                    dataset for dataset in datasets if dataset.dataset_type.type_id == self.dataset_type_id
                ][0]
            except IndexError:
                raise AirflowException(
                    f"No top level dataset type set, expecting dataset type with type_id: {self.dataset_type_id}"
                )
            releases = get_dataset_releases(dataset_id=top_level_dataset.id)
            latest_release = get_latest_dataset_release(releases=releases)
            start_date = pendulum.instance(latest_release.end_date).start_of("day")

        end_date = pendulum.instance(kwargs["data_interval_end"])
        logging.info(f"Start date: {start_date}, end date: {end_date}, first release: {first_release}")

        return start_date, end_date, first_release

    def get_bq_load_info(self, release: StreamRelease) -> List[Tuple[str, str, str]]:
        """Get (a list of) the transform blob, main table id and partition table id that are used to load data into
        BigQuery.
        When the batch_load property of the telescope is set to True, all blobs in the transform folder will be loaded
        into 1 table at once.
        When the batch_load property of the telescope is set to False, each blob in the transform folder will be used to
        create/update an individual table.

        :param release: The release object.
        :return: List with tuples of transform_blob, main_table_id and partition_table_id
        """
        bq_info = []
        if self.batch_load:
            if release.transform_files:
                transform_blob = batch_blob_name(release.transform_folder)
                main_table_id, partition_table_id = (self.dag_id, f"{self.dag_id}_partitions")
                bq_info = [(transform_blob, main_table_id, partition_table_id)]
        else:
            for transform_path in release.transform_files:
                transform_blob = blob_name(transform_path)
                main_table_id, partition_table_id = table_ids_from_path(transform_path)
                bq_info.append((transform_blob, main_table_id, partition_table_id))
        bq_info.sort(key=lambda x: x[0])
        return bq_info

    def download(self, release: StreamRelease, **kwargs):
        """Task to download the StreamRelease release.

        :param release: a StreamRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.download()

    def upload_downloaded(self, release: StreamRelease, **kwargs):
        """Upload the downloaded files for the given release.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, release: StreamRelease, **kwargs):
        """Task to extract the StreamRelease release.

        :param release: a StreamRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.extract()

    def transform(self, release: StreamRelease, **kwargs):
        """Task to transform the StreamRelease release.

        :param release: a StreamRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.transform()

    def upload_transformed(self, release: StreamRelease, **kwargs):
        """Upload the transformed files for the given release.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        upload_files_from_list(release.transform_files, release.transform_bucket)

    def bq_load_partition(self, release: StreamRelease, **kwargs):
        """For each file listed in transform_files, create a time based partition in BigQuery

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        if release.first_release:
            logging.info("Skipped, because first release")
            return

        bq_load_info = self.get_bq_load_info(release)
        for transform_blob, main_table_id, partition_table_id in bq_load_info:
            table_description = self.table_descriptions.get(main_table_id, "")
            bq_load_ingestion_partition(
                self.schema_folder,
                release.end_date,
                transform_blob,
                self.dataset_id,
                main_table_id,
                partition_table_id,
                self.source_format,
                self.schema_prefix,
                self.schema_version,
                self.dataset_description,
                table_description=table_description,
                **self.load_bigquery_table_kwargs,
            )

    def bq_delete_old(self, release: StreamRelease, **kwargs):
        """Delete old rows from the 'main' table, based on rows that are in a partition of the 'partitions' table.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        if release.first_release:
            logging.info("Skipped, because first release")
            return

        logging.info(f"Deleting old data from main table using partition with ingestion date of {release.end_date}")
        bytes_budget = kwargs.get("bytes_budget", None)
        bq_load_info = self.get_bq_load_info(release)
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        for _, main_table_id, partition_table_id in bq_load_info:
            bq_delete_old(
                release.end_date,
                self.dataset_id,
                main_table_id,
                partition_table_id,
                self.merge_partition_field,
                bytes_budget=bytes_budget,
                project_id=project_id,
            )

    def bq_append_new(self, release: StreamRelease, **kwargs):
        """Append rows to the 'main' table, based on rows that are in a partition of the 'partitions' table.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        bq_load_info = self.get_bq_load_info(release)

        if release.first_release:
            for transform_blob, main_table_id, partition_table_id in bq_load_info:
                table_description = self.table_descriptions.get(main_table_id, "")
                bq_append_from_file(
                    self.schema_folder,
                    release.end_date,
                    transform_blob,
                    self.dataset_id,
                    main_table_id,
                    self.source_format,
                    self.schema_prefix,
                    self.schema_version,
                    self.dataset_description,
                    table_description=table_description,
                    **self.load_bigquery_table_kwargs,
                )
            return

        logging.info(f"Appending data to main table from partition with ingestion date of {release.end_date}")
        for transform_blob, main_table_id, partition_table_id in bq_load_info:
            bq_append_from_partition(
                release.end_date,
                self.dataset_id,
                main_table_id,
                partition_table_id,
            )

    def bq_create_snapshot(self, release: StreamRelease, **kwargs):
        """Create a snapshot from an existing BigQuery table

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        logging.info("Creating a BigQuery snapshot")
        bq_load_info = self.get_bq_load_info(release)
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        for _, main_table_id, _ in bq_load_info:
            destination_table_id = bigquery_sharded_table_id(main_table_id + "_snapshots", release.end_date)
            create_bigquery_snapshot(project_id, self.dataset_id, main_table_id, self.dataset_id, destination_table_id)

    def cleanup(self, release: StreamRelease, **kwargs):
        """Delete downloaded, extracted and transformed files of the release. Deletes old xcoms.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.cleanup()

        execution_date = kwargs["execution_date"]
        delete_old_xcoms(dag_id=self.dag_id, execution_date=execution_date)

    def merge_update_files(self, release: StreamRelease, **kwargs):
        """Merge the transformed jsonl into a single jsonl and delete the parts.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """

        # Skip on first release.
        if release.first_release:
            return

        key = self.merge_partition_field
        input_files = release.transform_files
        input_files.sort()  # Assume lexicological order is correct order to apply updates.
        output_file = os.path.join(release.transform_folder, f"{self.DAG_ID}.jsonl")

        # Create mapping
        update_data = {}
        for i, input_file in enumerate(input_files):
            with open(input_file, "r") as f:
                for line in f:
                    row = json.loads(line)
                    update_data[row[key]] = i

        # Write out merged file
        with open(output_file, "w") as outf:
            for i, input_file in enumerate(input_files):
                with open(input_file, "r") as inf:
                    for line in inf:
                        row = json.loads(line)
                        if update_data[row[key]] == i:
                            outf.write(line)

        # Delete original parts
        for file in input_files:
            pathlib.Path(file).unlink()

        # Upload merged file to bucket.
        # Ideally in own task, but would it need branching mechanic (currently not implemented in Workflow).
        upload_files_from_list(release.transform_files, release.transform_bucket)



# Copyright 2018 Elasticsearch BV
# Copyright 2020, 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Some of the docstrings are based on the https://github.com/elastic/kibana/tree/7.10/docs/api documentation which is
# licensed under the Apache 2.0 license according to the license notice:
# https://github.com/elastic/kibana/blob/7.10/LICENSE.txt

# Author: James Diprose, Aniek Roelofs

import base64
import dataclasses
import json
import logging
from enum import Enum
from posixpath import join as urljoin
from typing import Dict, List

import requests


@dataclasses.dataclass
class TimeField:
    """A Kibana time field.

    :param pattern: a regex pattern. If the regex pattern matches an index, the associated time field name will be applied.
    :param field_name: the name of the Kibana time field.
    """

    pattern: str = None
    field_name: str = None


class ObjectType(Enum):
    """Valid Kibana saved object types"""

    visualization = "visualization"
    dashboard = "dashboard"
    search = "search"
    index_pattern = "index-pattern"
    config = "config"
    timelion_sheet = "timelion-sheet"


class Kibana:
    HTTP_NOT_FOUND = 404

    def __init__(
        self,
        host: str = "http://kibana:5601/",
        username: str = None,
        password: str = None,
        api_key_id: str = None,
        api_key: str = None,
        headers: dict = None,
    ):
        """Create a Kibana API client.

        :param host: the host including the hostname and port.
        :param username: the Kibana username
        :param password: the Kibana password
        :param api_key_id: the Kibana API key id.
        :param api_key: the Kibana API key.
        :param headers: the headers that will be used with the Kibana API
        """
        self.headers = headers
        if not self.headers:
            self.headers = {"Content-Type": "application/json", "kbn-xsrf": "true"}

        self.host = host

        if username and password:
            auth = base64.b64encode(f"{username}:{password}".encode()).decode()
            self.headers["Authorization"] = f"Basic {auth}"

        if api_key_id and api_key:
            auth = base64.b64encode(f"{api_key_id}:{api_key}".encode()).decode()
            self.headers["Authorization"] = f"ApiKey {auth}"

    def create_space(
        self,
        space_id: str,
        name: str,
        description: str = None,
        disabled_features: List = None,
        initials: str = None,
        color: str = None,
        image_url: str = None,
    ) -> bool:
        """Create a Kibana space.

        :param space_id: the space ID.
        :param name: the display name for the space.
        :param description: the description for the space.
        :param disabled_features: the list of disabled features for the space.
        :param initials: the initials shown in the space avatar. By default, the initials are automatically generated
        from the space name. Initials must be 1 or 2 characters.
        :param color: the hexadecimal color code used in the space avatar. By default, the color is automatically
        generated from the space name.
        :param image_url: The data-URL encoded image to display in the space avatar. If specified, initials will not
        be displayed, and the color will be visible as the background color for transparent images. For best results,
        your image should be 64x64. Images will not be optimized by this API call, so care should be taken when using
        custom images.
        :return: whether the Kibana space was created successfully or not.
        """

        # Construct body
        body = {"id": space_id, "name": name}
        if description is not None:
            body["description"] = description
        if disabled_features is not None:
            body["disabledFeatures"] = disabled_features
        if initials is not None:
            body["initials"] = initials
        if color is not None:
            body["color"] = color
        if image_url is not None:
            body["imageUrl"] = image_url

        url = self._make_spaces_url()
        response = requests.post(url, headers=self.headers, data=json.dumps(body))

        success = response.status_code == 200
        if not success:
            logging.error(response.text)

        return success

    def delete_space(self, space_id: str) -> bool:
        """Delete a Kibana space.

        :param space_id: the space ID.
        :return: whether the Kibana space was deleted successfully or not.
        """

        url = self._make_spaces_url(space_id=space_id)
        response = requests.delete(url, headers=self.headers)

        success = response.status_code == 200
        if not success:
            logging.error(response.text)

        return success

    def create_object(
        self,
        object_type: ObjectType,
        object_id: str,
        attributes: Dict,
        space_id: str = None,
        overwrite: bool = False,
        initial_namespaces: List[str] = None,
        exists_ok: bool = False,
    ) -> bool:
        """Create a Kibana saved object.

        :param object_type: the type of object to create.
        :param object_id: the object ID.
        :param attributes: the data that you want to create.
        :param space_id: the space ID. If space_id is not provided the default space is used.
        :param overwrite: when true, overwrites the document with the same ID.
        :param initial_namespaces: identifiers for the spaces in which this object is created. If this is provided, the
        object is created only in the explicitly defined spaces. If this is not provided, the object is created in the
        current space (default behavior).
        :param exists_ok: if the object already exists then consider creation successful.
        :return: whether the Kibana saved object was created successfully or not.
        """

        body = {"attributes": attributes}

        if initial_namespaces is not None:
            body["initialNamespaces"] = initial_namespaces

        params = (("overwrite", overwrite),)

        url = self._make_saved_object_url(object_type, object_id, space_id=space_id)
        response = requests.post(url, headers=self.headers, params=params, data=json.dumps(body))

        success = response.status_code == 200 or (exists_ok and response.status_code == 409)
        if not success:
            logging.error(response.text)

        return success

    def delete_object(self, object_type: ObjectType, object_id: str, space_id: str = None, force: bool = False) -> bool:
        """Delete a Kibana saved object.

        :param object_type: the object type.
        :param object_id: The object ID that you want to delete.
        :param space_id: the space ID. If not provided the default space is used.
        :param force: when true, forces an object to be deleted if it exists in multiple namespaces.
        :return: whether Kibana object was deleted successfully or not.
        """

        url = self._make_saved_object_url(object_type, object_id, space_id=space_id)

        params = (("force", force),)

        response = requests.delete(url, headers=self.headers, params=params)

        success = response.status_code == 200
        if not success:
            logging.error(response.text)

        return success

    def _make_spaces_url(self, space_id: str = None) -> str:
        """Make a URL for using the Kibana spaces REST API.

        :param space_id: the space ID.
        :return: the URL.
        """

        parts = ["api", "spaces", "space"]
        if space_id is not None:
            parts.append(space_id)
        return urljoin(self.host, *parts)

    def _make_saved_object_url(self, object_type: ObjectType, object_id: str, space_id: str = None) -> str:
        """Make a URL for using the Kibana saved object API.

        :param object_type: the Kibana saved object type.
        :param object_id: the object ID.
        :param space_id: the space ID.
        :return: the URL.
        """

        # Make URL
        parts = []
        if space_id is not None:
            parts += ["s", space_id]
        parts += ["api", "saved_objects", object_type.value, object_id]
        return urljoin(self.host, *parts)

    def _make_index_pattern_url(self, index_pattern_id: str, space_id: str = None):
        """Make an index pattern URL.

        :param index_pattern_id: the id of the index pattern.
        :param space_id: the Kibana space id.
        :return: the URL.
        """

        parts = []
        if space_id is not None:
            parts += ["s", space_id]
        parts += ["api", "index_patterns", "index_pattern", index_pattern_id]
        return urljoin(self.host, *parts)

    def get_index_pattern(self, index_pattern_id: str, space_id: str = None):
        """Get an index pattern.

        :param index_pattern_id: the id of the index pattern.
        :param space_id: the Kibana space id.
        :return: the index pattern details.
        """

        url = self._make_index_pattern_url(index_pattern_id, space_id=space_id)
        response = requests.get(url, headers=self.headers)

        # If index pattern is not found (404) return None
        if response.status_code == self.HTTP_NOT_FOUND:
            logging.error(response.text)
            return None

        return response.text



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import pendulum
from observatory.platform.utils.api import make_observatory_api
from observatory.platform.utils.config_utils import find_schema
from typing import List, Union, Tuple
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.api.client.model.dataset import Dataset
import datetime


def address_to_gcp_fields(address: str) -> Tuple[str, str, str]:
    """Take an address in the DatasetStorage field, and return the project ID, dataset ID, table ID.

    :param address: Address string in the DatasetStorage field.
    :return GCP project, dataset, table.
    """

    return tuple(address.split("."))


def get_datasets(*, workflow_id: int) -> List["Dataset"]:  # noqa: F821
    """Get a list of datasets for a workflow.
    Only returns the first 1000 datasets for that workflow.

    :param workflow_id: Workflow type ID.
    :return: List of Dataset objects.
    """

    limit = 1000
    api = make_observatory_api()
    datasets = api.get_datasets(workflow_id=workflow_id, limit=limit)
    return datasets


def get_dataset_releases(*, dataset_id: int) -> List[DatasetRelease]:
    """Get a list of dataset releases for a given dataset.

    :param dataset_id: Dataset id.
    :return: List of dataset releases.
    """

    limit = 1000
    api = make_observatory_api()
    dataset_releases = api.get_dataset_releases(dataset_id=dataset_id, limit=limit)
    return dataset_releases


def get_latest_dataset_release(releases: List[DatasetRelease]) -> DatasetRelease:
    """Get the dataset release from the list with the most recent end date.

    :param releases: List of releases.
    :return: Latest release (by end_date)
    """

    latest = releases[0]
    for release in releases:
        if release.end_date > latest.end_date:
            latest = release

    return latest


def get_start_end_date(
    release: "AbstractRelease",  # noqa: F821
) -> Tuple[pendulum.DateTime, pendulum.DateTime]:
    """Get the start date and end date from a release object.
    Supports start_date, end_date, release_date, release_id attributes.
    Those attributes need to be date or datetime objects.

    :param release: Release object.
    :return: Start date, end date.
    """

    if "start_date" in release.__dict__ and "end_date" in release.__dict__:
        start_date = release.start_date
        end_date = release.end_date
    elif "release_date" in release.__dict__:
        start_date = release.release_date
        end_date = release.release_date
    else:  # "release_id" in release.__dict__:  # This will only work if release_id is a date.
        start_date = pendulum.parse(release.release_id)
        end_date = pendulum.parse(release.release_id)

    return start_date, end_date


def get_new_release_dates(*, dataset_id: int, releases: List[datetime.datetime]) -> List[str]:
    """Get a list of new release dates, i.e., releases in the user supplied list that are not in the API db.

    :param dataset_id: The dataset id to check.
    :param releases: List of release dates to check.
    :return: List of new releases in YYYYMMDD string format.
    """

    api_releases = get_dataset_releases(dataset_id=dataset_id)
    api_releases_set = set([release.end_date.strftime("%Y%m%d") for release in api_releases])
    releases_set = set([release.strftime("%Y%m%d") for release in releases])
    new_releases = list(releases_set.difference(api_releases_set))
    return new_releases


def get_start_end_date_key(release: Tuple[datetime.datetime, datetime.datetime]) -> str:
    """Construct a string key from a start and end date.

    :param release: A (start date, end date) pair.
    :return: Key string.
    """

    key = release[0].strftime("%Y%m%d") + release[1].strftime("%Y%m%d")
    return key


def get_new_start_end_release_dates(
    *, dataset_id: int, releases: List[Tuple[datetime.datetime, datetime.datetime]]
) -> List[Tuple[str, str]]:
    """Get a list of (start date, end date) that are in the list supplied by the user but not in the API db.

    :param dataset_id: The dataset id to check.
    :param releases: List of (start date, end date) to check.
    :return: List of new (start date, end date) in YYYYMMDD string format.
    """

    api_releases = get_dataset_releases(dataset_id=dataset_id)
    api_releases_set = set([get_start_end_date_key((release.start_date, release.end_date)) for release in api_releases])
    releases_set = set([get_start_end_date_key(release) for release in releases])
    new_releases = releases_set.difference(api_releases_set)

    output = [(release[:8], release[8:]) for release in new_releases]
    return output


def is_first_release(workflow_id: int) -> bool:
    """Use the API to check whether this is the first release of a dataset, i.e., are there no dataset release records.

    :param workflow_id: API workflow id.
    :return: Whether this is the first release.
    """

    datasets = get_datasets(workflow_id=workflow_id)
    releases = get_dataset_releases(dataset_id=datasets[0].id)
    return len(releases) == 0


def add_dataset_release(
    *,
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime,
    dataset_id: int,
):
    """Add a new dataset release record to the API database.

    :param start_date: Start date.
    :param end_date: End date.
    :param dataset_id: API Dataset id.
    """

    dataset_release = DatasetRelease(
        start_date=start_date,
        end_date=end_date,
        dataset=Dataset(id=dataset_id),
    )
    api = make_observatory_api()
    api.post_dataset_release(dataset_release)



# Copyright 2019 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien

# Asynchronous HTTP GET file downloader. Use the download_file, and download_files interfaces to download.
# Creates a worker pool to asynchronously (single thread) download file(s).
# Valid hash algorithms: see observatory.platform.utils.file_utils.get_hasher_ for valid options.
#
# Usage examples:
#    custom_headers = {"User-Agent" : "Something" }
#  - download_file(url=url)
#  - download_file(url=url, filename="destination", headers=custom_headers)
#  - download_file(url=url, filename="destination", hash="d733dac9083babc757972353d75401ed", hash_algorithm="md5")
#
#  - download_files(download_list=["http://url1/file", "http://url2/file"])
#    download_list=[
#      {"url": "http://myurl/file", "filename": "dst_file.txt"},
#      {"url": "http://myurl2/file", "filename": "dst_file2.txt", "hash": "d733dac9083babc757972353d75401ed", "hash_algorithm" : "md5"},
#    ]
#  - download_files(download_list=download_list, num_connections=4, headers=custom_headers)
#
# Supports:
#  - Retrying downloads a few times in case of temporary failures.
#  - Retrying once if a supplied hash does not match the download.
#  - Skipping downloads if the file exists, and the hash matches the one supplied.

import asyncio
import logging
import os
from cgi import parse_header
from dataclasses import dataclass
from queue import Queue
from typing import Any, Dict, List, Union

import aiohttp
import backoff
import validators
from observatory.platform.utils.file_utils import validate_file_hash
from observatory.platform.utils.url_utils import get_filename_from_url


@dataclass
class DownloadInfo:
    """Metadata needed to download a file."""

    url: str  # URL to download
    filename: Union[str, None] = None  # Destination file
    hash: Union[str, None] = None  # Hash code for file integrity checking
    hash_algorithm: Union[str, None] = None  # Hash algorithm for the hash
    prefix_dir: str = ""  # Prefix the filename path with this
    retry: bool = False  # Whether to retry download


@backoff.on_exception(backoff.expo, aiohttp.ClientError, max_tries=11, max_time=60)
async def download_http_file_(*, download_info: DownloadInfo, headers=None):
    """Download a single file from a HTTP GET request.

    :param download_info: Download information.
    :param headers: Optional headers to use when making get request, e.g., if providing custom User Agent.
    """

    READ_BUFFER_SIZE = 2 ** 16  # 64 KiB

    url = download_info.url
    dst_file = download_info.filename

    async with aiohttp.ClientSession(raise_for_status=True, headers=headers) as session:
        async with session.get(url, timeout=None) as resp:
            if dst_file is None:
                _, params = parse_header(resp.headers.get("Content-Disposition", ""))
                if params != "" and "filename" in params:
                    dst_file = params["filename"]
                else:
                    dst_file = get_filename_from_url(url=url)

                download_info.filename = dst_file
            dst_file = os.path.join(download_info.prefix_dir, dst_file)

            logging.info(f"downloading {url} to {dst_file}")

            with open(dst_file, "wb") as f:
                while True:
                    chunk = await resp.content.read(READ_BUFFER_SIZE)
                    if not chunk:
                        break
                    f.write(chunk)


def skip_download(*, download_info: Dict) -> bool:
    """Check whether we should skip the download.  It will check if the file exists and its hash matches a supplied
    hash. If no hash is supplied, we will not skip.

    :param download_info: Download job information.
    :return: Whether we should skip the download for this file.
    """

    hash = download_info.hash
    hash_algorithm = download_info.hash_algorithm
    prefix_dir = download_info.prefix_dir
    filename = download_info.filename

    if filename is None:
        return False

    if prefix_dir != "":
        filename = os.path.join(prefix_dir, filename)

    if hash_algorithm is not None and os.path.exists(filename):
        valid = validate_file_hash(file_path=filename, expected_hash=hash, algorithm=hash_algorithm)
        if valid:
            logging.info(f"File {filename} exists, and has a valid hash file. Skipping download.")
            return True
        else:
            logging.info(f"File {filename} exists, but has a bad hash. Redownloading.")

    return False


def requeue_once_if_bad_hash(
    *, download_info: DownloadInfo, exception: Union[None, Exception], downloads: asyncio.Queue, errors: List[Exception]
):
    """If a hash was supplied, it checks if the file downloaded passes hash validation. If it doesn't, then on the
    first failure, it will requeue the download.  Otherwise, it will produce an error message and give up on that
    download.  If an exception was raised during download, it will not requeue.

    :param download_info: Download job information.
    :param exception: Whether an exception was raised during download.
    :param downloads: Queue of download job information.
    :param errors: Queue of encountered error objects for later processing.
    """

    hash = download_info.hash
    hash_algorithm = download_info.hash_algorithm
    prefix_dir = download_info.prefix_dir
    filename = download_info.filename

    if filename is None:
        return

    if prefix_dir != "":
        filename = os.path.join(prefix_dir, filename)

    if hash_algorithm is not None and exception is None and os.path.exists(filename):
        valid = validate_file_hash(file_path=filename, expected_hash=hash, algorithm=hash_algorithm)

        # Requeue download the first time there's a bad hash.
        if not valid and download_info.retry == False:
            warning_msg = f"File {filename} and hash {hash} do not match. Retrying download."
            logging.warning(warning_msg)
            download_info.retry = True
            downloads.put_nowait(download_info)

        # Error out if this is not the first attempt.
        elif not valid and download_info.retry:
            error_msg = f"File {filename} has a bad hash after retrying download. Giving up."
            errors.append(Exception(error_msg))


async def worker_(name: str, downloads: asyncio.Queue, errors: List[Exception], headers: Union[None, Dict]):
    """Worker that fetches download jobs and executes downloads.

    :param name: Name of the worker.
    :param downloads: Download queue of jobs.
    :param errors: List of encountered download errors.
    :param headers: Custom HTTP headers to use.
    """

    logging.info(f"Starting worker {name}")
    while True:
        # If there are errors, terminate the worker thread.
        if len(errors) > 0:
            break

        download_info = await downloads.get()

        # Skip if file exists and has good hash
        if skip_download(download_info=download_info):
            downloads.task_done()
            continue

        # Download the file
        url = download_info.url
        logging.info(f"{name}: downloading {url}")

        exception = None
        try:
            await download_http_file_(download_info=download_info, headers=headers)
        except Exception as e:
            errors.append(e)
            exception = e

        # Requeue on bad hash once.
        requeue_once_if_bad_hash(download_info=download_info, exception=exception, downloads=downloads, errors=errors)

        downloads.task_done()


async def download_http_files_(
    *,
    download_list: List[DownloadInfo],
    num_connections: str = 8,
    headers: Dict = None,
) -> bool:
    """Download a list of files via HTTP asynchronously.  Supports multiple connections
    and custom headers.

    :param download_list: List of files to download.
    :param num_connections: Maximum number of concurrent connections to use.
    :param headers: Custom headers to use in HTTP GET request.
    :return: True on success, False on failure.
    """

    downloads = asyncio.Queue()
    errors = list()

    # Load jobs in queue
    for task in download_list:
        downloads.put_nowait(task)

    # Create workers
    workers = list()
    num_workers = min(num_connections, len(download_list))
    for i in range(num_workers):
        name = f"{i}"
        worker = asyncio.create_task(worker_(name, downloads, errors, headers))
        workers.append(worker)

    # Block until all jobs done
    await downloads.join()

    # Stop all workers
    for worker in workers:
        worker.cancel()

    # Block until all workers are cancelled.
    await asyncio.gather(*workers, return_exceptions=True)

    for error in errors:
        logging.error(error)

    success = len(errors) == 0
    return success


def download_files(
    *,
    download_list: List[Union[str, DownloadInfo]],
    num_connections: int = 8,
    headers: Dict = None,
    prefix_dir: str = "",
) -> bool:
    """Download a list of files. Can support simultaneous connections and custom HTTP headers.

    :param download_list: List of download jobs. You can specify a list of url strings, or a dictionary of the form
    {"url": "urlstring", "filename": "savedfile", "hash" : "hashcode", "hash_algorithm": "hash algorithm"}
    :param num_connections: Maximum number of concurrent connections.
    :param headers: Custom HTTP header to use for downloading.
    :param prefix_dir: A directory to prefix the filename path with. If specified and the download_list is a list of DownloadInfo, it overrides the prefix_dir in each DownloadInfo.
    :return: True on sucess, False on failure.
    """

    if len(download_list) == 0:
        return

    # Convert list of urls to download info dict.
    for i, info in enumerate(download_list):
        if isinstance(info, str):
            info = DownloadInfo(url=info, prefix_dir=prefix_dir)
            download_list[i] = info
        elif isinstance(info, DownloadInfo):
            if prefix_dir != "":
                info.prefix_dir = prefix_dir
        else:
            raise Exception(f"Expecting a DownloadInfo object. Received a {type(info)}.")

    success = asyncio.run(
        download_http_files_(download_list=download_list, num_connections=num_connections, headers=headers),
        debug=True,
    )
    return success


def download_file(
    *,
    url: str,
    filename: str = None,
    headers: Dict = None,
    hash: str = None,
    hash_algorithm: str = None,
    prefix_dir: str = "",
) -> bool:
    """Download a single file from a url.

    :param url: URL to download file from.
    :param filename: Destination file.
    :param headers: Any custom header you want to use in HTTP session.
    :param prefix_dir: A directory to prefix the filename path with.
    :return: True on sucess, False on failure.
    """

    download_info = DownloadInfo(
        url=url,
        filename=filename,
        prefix_dir=prefix_dir,
    )

    if hash_algorithm is not None:
        download_info.hash = hash
        download_info.hash_algorithm = hash_algorithm

    success = download_files(download_list=[download_info], num_connections=1, headers=headers)
    return success



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose, Tuan Chien

from typing import Dict, List, Optional

import pendulum
from google.cloud.bigquery import SourceFormat
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.workflow_utils import (
    blob_name,
    bq_load_shard,
    delete_old_xcoms,
    table_ids_from_path,
    upload_files_from_list,
)
from observatory.platform.workflows.workflow import Release, Workflow
from sqlalchemy.sql.expression import delete


class SnapshotRelease(Release):
    def __init__(
        self,
        dag_id: str,
        release_date: pendulum.DateTime,
        download_files_regex: str = None,
        extract_files_regex: str = None,
        transform_files_regex: str = None,
    ):
        """Construct a SnapshotRelease instance.

        :param dag_id: the id of the DAG.
        :param release_date: the release date (used to construct release_id).
        :param download_files_regex: regex pattern that is used to find files in download folder
        :param extract_files_regex: regex pattern that is used to find files in extract folder
        :param transform_files_regex: regex pattern that is used to find files in transform folder
        """
        self.release_date = release_date
        release_id = f'{dag_id}_{self.release_date.strftime("%Y_%m_%d")}'
        super().__init__(dag_id, release_id, download_files_regex, extract_files_regex, transform_files_regex)


class SnapshotTelescope(Workflow):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        schedule_interval: str,
        dataset_id: str,
        schema_folder: str,
        catchup: bool = True,
        queue: str = "default",
        max_retries: int = 3,
        max_active_runs: int = 1,
        source_format: str = SourceFormat.NEWLINE_DELIMITED_JSON,
        schema_prefix: str = "",
        schema_version: str = None,
        load_bigquery_table_kwargs: Dict = None,
        dataset_description: str = "",
        table_descriptions: Dict[str, str] = None,
        airflow_vars: list = None,
        airflow_conns: list = None,
        workflow_id: int = None,
        tags: Optional[List[str]] = None,
        **kwargs,
    ):
        """Construct a SnapshotTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the dataset id.
        :param schema_folder: the path to the SQL schema folder.
        :param catchup: whether to catchup the DAG or not.
        :param queue: the Airflow queue name.
        :param max_retries: the number of times to retry each task.
        :param max_active_runs: the maximum number of DAG runs that can be run at once.
        :param source_format: the format of the data to load into BigQuery.
        :param schema_prefix: the prefix used to find the schema path.
        :param schema_version: the version used to find the schema path.
        :param load_bigquery_table_kwargs: the customisation parameters for loading data into a BigQuery table.
        :param dataset_description: description for the BigQuery dataset.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        :param airflow_conns: list of airflow connection keys, for each connection it is checked if it exists in airflow
        :param workflow_id: api workflow id.
        :param tags: Optional Airflow DAG tags to add.
        """

        # Set transform_bucket_name as required airflow variable
        if not airflow_vars:
            airflow_vars = []
        airflow_vars = list(set([AirflowVars.TRANSFORM_BUCKET] + airflow_vars))
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            catchup,
            queue,
            max_retries,
            max_active_runs,
            airflow_vars,
            airflow_conns,
            workflow_id=workflow_id,
            tags=tags,
            **kwargs,
        )
        self.dataset_id = dataset_id
        self.schema_folder = schema_folder
        self.source_format = source_format
        self.schema_prefix = schema_prefix
        self.schema_version = schema_version
        self.load_bigquery_table_kwargs = load_bigquery_table_kwargs if load_bigquery_table_kwargs else dict()
        self.dataset_description = dataset_description
        self.table_descriptions = table_descriptions if table_descriptions else dict()

    def download(self, releases: List[SnapshotRelease], **kwargs):
        """Task to download the releases.

        :param releases: a list of SnapshotRelease.
        :return: None.
        """
        for release in releases:
            release.download()

    def upload_downloaded(self, releases: List[SnapshotRelease], **kwargs):
        """Task to upload each downloaded release to a Google Cloud bucket.

        :param releases: a list of releases.
        :param kwargs: the context passed from the PythonOperator.
        :return: None.
        """
        for release in releases:
            upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, releases: List[SnapshotRelease], **kwargs):
        """Task to extract the releases.

        :param releases: a list of SnapshotRelease.
        :return: None.
        """
        for release in releases:
            release.extract()

    def transform(self, releases: List[SnapshotRelease], **kwargs):
        """Task to transform the releases.

        :param releases: A list of SnapshotRelease objects.
        :return: None.
        """
        for release in releases:
            release.transform()

    def upload_transformed(self, releases: List[SnapshotRelease], **kwargs):
        """Task to upload each transformed release to a Google Cloud bucket.

        :param releases: a list of releases.
        :param kwargs: the context passed from the PythonOperator.
        :return: None.
        """
        for release in releases:
            upload_files_from_list(release.transform_files, release.transform_bucket)

    def bq_load(self, releases: List[SnapshotRelease], **kwargs):
        """Task to load each transformed release to BigQuery.
        The table_id is set to the file name without the extension.

        :param releases: a list of releases.
        :param kwargs: the context passed from the PythonOperator.
        :return: None.
        """
        for release in releases:
            for transform_path in release.transform_files:
                transform_blob = blob_name(transform_path)
                table_id, _ = table_ids_from_path(transform_path)
                table_description = self.table_descriptions.get(table_id, "")
                bq_load_shard(
                    self.schema_folder,
                    release.release_date,
                    transform_blob,
                    self.dataset_id,
                    table_id,
                    self.source_format,
                    prefix=self.schema_prefix,
                    schema_version=self.schema_version,
                    dataset_description=self.dataset_description,
                    table_description=table_description,
                    **self.load_bigquery_table_kwargs,
                )

    def cleanup(self, releases: List[SnapshotRelease], **kwargs):
        """Delete files of downloaded, extracted and transformed release. Deletes old xcoms.

        :param releases: a list of releases.
        :param kwargs: the context passed from the PythonOperator.
        :return: None.
        """
        for release in releases:
            release.cleanup()

        execution_date = kwargs["execution_date"]
        delete_old_xcoms(dag_id=self.dag_id, execution_date=execution_date)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The keywords airflow and DAG are required to load the DAGs from this file, see bullet 2 in the Apache Airflow FAQ:
# https://airflow.apache.org/docs/stable/faq.html

# Author: James Diprose

import logging

from airflow.models import DagBag

from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.workflow_utils import fetch_dags_modules, fetch_dag_bag

# Load DAGs for each DAG path
dags_modules = fetch_dags_modules()
for module_name in dags_modules:
    dags_path = module_file_path(module_name)
    logging.info(f"{module_name} DAGs path: {dags_path}")
    dag_bag: DagBag = fetch_dag_bag(dags_path)

    # Load dags
    for dag_id, dag in dag_bag.dags.items():
        logging.info(f"Adding DAG: dag_id={dag_id}, dag={dag}")
        globals()[dag_id] = dag
